package tetris.model;

import tetris.Tetris;
import tetris.gui.ActionHandler;
import tetris.gui.GUI;
import tetris.model.figures.IFigure;
import tetris.model.figures.JFigure;
import tetris.model.figures.LFigure;
import tetris.model.figures.OFigure;
import tetris.model.figures.SFigure;
import tetris.model.figures.TFigure;
import tetris.model.figures.ZFigure;

import java.util.Random;

/**
 * The class Game implements the Tetris game.
 *
 * @author Stephan Fischli
 * @version 2.1
 */
public class Game {

	/**
	 * The graphical user interface.
	 */
	private final GUI gui;

	/*
	 * The playing field.
	 */
	private final Field field;

	/**
	 * The figure of the game.
	 */
	private Figure figure;

	/**
	 * Constructs a game with a graphical user interface.
	 *
	 * @param gui the graphical user interface
	 */
	public Game(GUI gui) {
		this.gui = gui;
		this.field = new Field();
	}

	/**
	 * Starts the game by creating a figure and registering an action handler.
	 */
	public void start() {
		createFigure();
		gui.setActionHandler(new FigureController());
	}

	/**
	 * Creates a random figure at the top of the field.
	 */
	private void createFigure() {
		int x = (Tetris.WIDTH - 1) / 2;
		int y = Tetris.HEIGHT - 1;
		int type = (int) (7 * Math.random());
		figure = switch (type) {
			case 0 -> new IFigure(x, y);
			case 1 -> new JFigure(x, y);
			case 2 -> new LFigure(x, y);
			case 3 -> new OFigure(x, y);
			case 4 -> new SFigure(x, y);
			case 5 -> new TFigure(x, y);
			default -> new ZFigure(x, y);
		};
		updateGUI();
	}

	/**
	 * Updates the graphical user interface according to the current state of the game.
	 */
	private void updateGUI() {
		gui.clear();
		gui.drawBlocks(figure.getBlocks());
	}

	/**
	 * The class FigureController is used to control the figure of the Tetris game.
	 *
	 * @author Stephan Fischli
	 * @version 2.1
	 */
	private class FigureController implements ActionHandler {

		/**
		 * Shifts the figure down.
		 */
		@Override
		public void shiftDown() throws CollisionException {
			figure.shift(0, -1);
			if (field.detectCollision(figure.getBlocks())) {
				figure.shift(0, 1);
				throw new CollisionException("Movement not possible");
			}
			updateGUI();
		}

		/**
		 * Shifts the figure left.
		 */
		@Override
		public void shiftLeft() throws CollisionException {
			figure.shift(-1, 0);
			if (field.detectCollision(figure.getBlocks())) {
				figure.shift(1, 0);
				throw new CollisionException("Movement not possible");
			}
			updateGUI();
		}

		/**
		 * Shifts the figure right.
		 */
		@Override
		public void shiftRight() throws CollisionException {
			figure.shift(1, 0);
			if (field.detectCollision(figure.getBlocks())) {
				figure.shift(-1, 0);
				throw new CollisionException("Movement not possible");
			}
			updateGUI();
		}

		/**
		 * Rotates the figure to the left.
		 */
		@Override
		public void rotateLeft() throws CollisionException {
			figure.rotate(-1);
			if (field.detectCollision(figure.getBlocks())) {
				figure.rotate(1);
				throw new CollisionException("Movement not possible");
			}
			updateGUI();
		}

		/**
		 * Rotates the figure to the right.
		 */
		@Override
		public void rotateRight() throws CollisionException {
			figure.rotate(1);
			if (field.detectCollision(figure.getBlocks())) {
				figure.rotate(-1);
				throw new CollisionException("Movement not possible");
			}
			updateGUI();
		}

		/**
		 * Drops the figure.
		 */
		@Override
		public void drop() {
			while (!field.detectCollision(figure.getBlocks())) {
				figure.shift(0, -1);
			}
			figure.shift(0, 1);
			updateGUI();
		}
	}
}
